"""Controller holding and managing Afero IoT resources of type `portable-air-conditioner`."""

from ... import device
from ...device import AferoDevice, AferoState
from ...errors import DeviceNotFound
from ...util import calculate_hubspace_celsius, process_function
from ..models import features
from ..models.portable_ac import PortableAC, PortableACPut
from ..models.resource import DeviceInformation, ResourceTypes
from .base import BaseResourcesController


class PortableACController(BaseResourcesController[PortableAC]):
    """Controller holding and managing Afero IoT resources of type `portable-air-conditioner`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.PORTABLE_AC]
    ITEM_CLS = PortableAC
    ITEM_MAPPING = {
        "hvac_mode": "mode",
    }
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {}
    # Elements that map to numbers. func class / func instance to unit
    ITEM_NUMBERS: dict[tuple[str, str | None], str] = {
        ("timer", None): "minutes",
    }
    # Elements that map to Select. func class / func instance to name
    ITEM_SELECTS = {
        ("fan-speed", "ac-fan-speed"): "Fan Speed",
        ("sleep", None): "Sleep Mode",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> PortableAC:
        """Initialize the element"""
        available: bool = False
        # Afero reports in Celsius by default
        display_celsius: bool = True
        current_temperature: features.CurrentTemperatureFeature | None = None
        hvac_mode: features.HVACModeFeature | None = None
        target_temperature_cooling: features.TargetTemperatureFeature | None = None
        numbers: dict[tuple[str, str], features.NumbersFeature] = {}
        selects: dict[tuple[str, str], features.SelectFeature] = {}
        for state in afero_device.states:
            func_def = device.get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if (
                state.functionClass == "temperature"
                and state.functionInstance == "current-temp"
            ):
                current_temperature = features.CurrentTemperatureFeature(
                    temperature=round(state.value, 1),
                    function_class=state.functionClass,
                    function_instance=state.functionInstance,
                )
            elif (
                state.functionClass == "temperature"
                and state.functionInstance == "cooling-target"
            ):
                target_temperature_cooling = generate_target_temp(
                    func_def["values"][0], state
                )

            elif state.functionClass == "mode":
                all_modes = set(process_function(afero_device.functions, "mode"))
                hvac_mode = features.HVACModeFeature(
                    mode=state.value,
                    previous_mode=state.value,
                    modes=all_modes,
                    supported_modes=all_modes,
                )
            elif state.functionClass == "available":
                available = state.value
            elif number := await self.initialize_number(func_def, state):
                numbers[number[0]] = number[1]
            elif select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]
            elif state.functionClass == "temperature-units":
                display_celsius = state.value == "celsius"

        self._items[afero_device.id] = PortableAC(
            afero_device.functions,
            id=afero_device.id,
            available=available,
            current_temperature=current_temperature,
            hvac_mode=hvac_mode,
            target_temperature_cooling=target_temperature_cooling,
            numbers=numbers,
            selects=selects,
            binary_sensors={},
            sensors={},
            display_celsius=display_celsius,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        updated_keys = set()
        cur_item = self.get_device(afero_device.id)
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add("available")
            elif (
                state.functionClass == "temperature"
                and state.functionInstance == "current-temp"
            ):
                if cur_item.current_temperature.temperature != round(state.value, 1):
                    cur_item.current_temperature.temperature = round(state.value, 1)
                    updated_keys.add(f"temperature-{state.functionInstance}")
            elif (
                state.functionClass == "temperature"
                and state.functionInstance == "cooling-target"
            ):
                if cur_item.target_temperature_cooling.value != state.value:
                    cur_item.target_temperature_cooling.value = state.value
                    updated_keys.add(f"temperature-{state.functionInstance}")
            elif state.functionClass == "mode":
                if cur_item.hvac_mode.mode != state.value:
                    cur_item.hvac_mode.previous_mode = cur_item.hvac_mode.mode
                    cur_item.hvac_mode.mode = state.value
                    updated_keys.add(state.functionClass)
            elif update_key := await self.update_number(state, cur_item):
                updated_keys.add(update_key)
            elif update_key := await self.update_select(state, cur_item):
                updated_keys.add(update_key)
            elif state.functionClass == "temperature-units":
                new_val: bool = state.value == "celsius"
                if cur_item.display_celsius != new_val:
                    cur_item.display_celsius = new_val
                    updated_keys.add(state.functionClass)
        return updated_keys

    async def set_state(self, device_id: str, **kwargs) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = PortableACPut()
        hvac_mode: str | None = kwargs.get("hvac_mode")
        target_temperature: float | None = kwargs.get("target_temperature")
        numbers: dict[tuple[str, str | None], float] | None = kwargs.get("numbers")
        selects: dict[tuple[str, str | None], str] | None = kwargs.get("selects")
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if hvac_mode:
            if hvac_mode in cur_item.hvac_mode.supported_modes:
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode=hvac_mode,
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown hvac mode %s. Available modes: %s",
                    hvac_mode,
                    ", ".join(sorted(cur_item.hvac_mode.modes)),
                )
        if target_temperature is not None:
            if not cur_item.display_celsius and not kwargs.get("is_celsius", False):
                target_temperature = calculate_hubspace_celsius(target_temperature)
            update_obj.target_temperature_cooling = features.TargetTemperatureFeature(
                value=target_temperature,
                min=cur_item.target_temperature_cooling.min,
                max=cur_item.target_temperature_cooling.max,
                step=cur_item.target_temperature_cooling.step,
                instance=cur_item.target_temperature_cooling.instance,
            )
        if numbers:
            for key, val in numbers.items():
                if key not in cur_item.numbers:
                    continue
                update_obj.numbers[key] = features.NumbersFeature(
                    value=val,
                    min=cur_item.numbers[key].min,
                    max=cur_item.numbers[key].max,
                    step=cur_item.numbers[key].step,
                    name=cur_item.numbers[key].name,
                    unit=cur_item.numbers[key].unit,
                )
                # Currently only ("timer", None) exists
                update_obj.current_temperature = features.CurrentTemperatureFeature(
                    temperature=cur_item.current_temperature.temperature + 1,
                    function_class=cur_item.current_temperature.function_class,
                    function_instance=cur_item.current_temperature.function_instance,
                )
        if selects:
            for key, val in selects.items():
                if key not in cur_item.selects:
                    continue
                update_obj.selects[key] = features.SelectFeature(
                    selected=val,
                    selects=cur_item.selects[key].selects,
                    name=cur_item.selects[key].name,
                )
        await self.update(device_id, obj_in=update_obj)


def generate_target_temp(
    func_def: dict, state: AferoState
) -> features.TargetTemperatureFeature:
    return features.TargetTemperatureFeature(
        value=round(state.value, 2),
        step=func_def["range"]["step"],
        min=func_def["range"]["min"],
        max=func_def["range"]["max"],
        instance=state.functionInstance,
    )



from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor


@dataclass
class Switch:
    """Representation of an Afero Switch"""

    id: str  # ID used when interacting with Afero
    available: bool

    on: dict[str | None, features.OnFeature]
    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    sensors: dict[str, AferoSensor] = field(default_factory=lambda: dict())
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=lambda: dict())

    type: ResourceTypes = ResourceTypes.FAN

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)


@dataclass
class SwitchPut:
    """States that can be updated for a Switch"""

    on: features.OnFeature | None = None



__all__ = [
    "Device",
    "DeviceInformation",
    "Light",
    "Lock",
    "AferoSensor",
    "AferoBinarySensor",
    "Switch",
    "Valve",
    "Fan",
    "ResourceTypes",
    "Thermostat",
    "ExhaustFan",
    "ExhaustFanPut",
    "FanPut",
    "LightPut",
    "LockPut",
    "SwitchPut",
    "ThermostatPut",
    "ValvePut",
    "PortableAC",
    "PortableACPut",
]


from .device import Device, DeviceInformation
from .exhaust_fan import ExhaustFan, ExhaustFanPut
from .fan import Fan, FanPut
from .light import Light, LightPut
from .lock import Lock, LockPut
from .portable_ac import PortableAC, PortableACPut
from .resource import ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor
from .switch import Switch, SwitchPut
from .thermostat import Thermostat, ThermostatPut
from .valve import Valve, ValvePut



from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor


@dataclass
class ExhaustFan:
    """Representation of an Afero Exhaust Fan"""

    id: str  # ID used when interacting with Afero
    available: bool

    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None
    selects: dict[tuple[str, str | None], features.SelectFeature] | None
    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    sensors: dict[str, AferoSensor] = field(default_factory=lambda: dict())
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=lambda: dict())

    type: ResourceTypes = ResourceTypes.EXHAUST_FAN

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        instances = {}
        for function in functions:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances


@dataclass
class ExhaustFanPut:
    """States that can be updated for an exhaust fan"""

    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=lambda: dict(), repr=False, init=False
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=lambda: dict(), repr=False, init=False
    )



from dataclasses import dataclass, field

from ...util import calculate_hubspace_fahrenheit
from ..models import features
from .resource import DeviceInformation, ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor


@dataclass
class PortableAC:
    """Representation of an Afero Portable AC"""

    id: str  # ID used when interacting with Afero
    available: bool

    display_celsius: bool | None
    current_temperature: features.CurrentTemperatureFeature | None
    hvac_mode: features.HVACModeFeature | None
    target_temperature_cooling: features.TargetTemperatureFeature | None
    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None
    selects: dict[tuple[str, str | None], features.SelectFeature] | None

    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    sensors: dict[str, AferoSensor] = field(default_factory=lambda: dict())
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=lambda: dict())

    type: ResourceTypes = ResourceTypes.PORTABLE_AC

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances

    @property
    def target_temperature(self) -> float | None:
        if self.display_celsius:
            return self.target_temperature_cooling.value
        else:
            return calculate_hubspace_fahrenheit(self.target_temperature_cooling.value)

    @property
    def target_temperature_step(self) -> float:
        if self.display_celsius:
            return self.target_temperature_cooling.step
        else:
            return 1

    @property
    def target_temperature_max(self) -> float:
        if self.display_celsius:
            return self.target_temperature_cooling.max
        else:
            return calculate_hubspace_fahrenheit(self.target_temperature_cooling.max)

    @property
    def target_temperature_min(self) -> float | None:
        if self.display_celsius:
            return self.target_temperature_cooling.min
        else:
            return calculate_hubspace_fahrenheit(self.target_temperature_cooling.min)

    @property
    def supports_fan_mode(self) -> bool:
        return False

    @property
    def supports_temperature_range(self) -> bool:
        return False

    @property
    def temperature(self) -> float | None:
        if self.display_celsius:
            return self.current_temperature.temperature
        else:
            return calculate_hubspace_fahrenheit(self.current_temperature.temperature)

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)


@dataclass
class PortableACPut:
    """States that can be updated for a Thermostat"""

    # This feels wrong but based on data dumps, setting timer increases the
    # current temperature by 1 to turn it on
    current_temperature: features.CurrentTemperatureFeature | None = None
    hvac_mode: features.HVACModeFeature | None = None
    target_temperature_cooling: features.TargetTemperatureFeature | None = None
    numbers: dict[tuple[str, str | None], features.NumbersFeature] | None = field(
        default_factory=lambda: dict(), repr=False, init=False
    )
    selects: dict[tuple[str, str | None], features.SelectFeature] | None = field(
        default_factory=lambda: dict(), repr=False, init=False
    )



"""Test Portable ACs"""

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers.portable_ac import PortableACController, features

from .. import utils

portable_ac = utils.create_devices_from_data("portable-ac.json")[0]
portable_ac_id = "8d0414d6-a7f7-4bdb-99d5-d866318ff559"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = PortableACController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == portable_ac_id
    assert dev.available is True
    assert dev.current_temperature == features.CurrentTemperatureFeature(
        temperature=35,
        function_class="temperature",
        function_instance="current-temp",
    )
    assert dev.hvac_mode == features.HVACModeFeature(
        mode="auto-cool",
        previous_mode="auto-cool",
        modes={"fan", "auto-cool", "dehumidify", "cool"},
        supported_modes={"fan", "auto-cool", "dehumidify", "cool"},
    )
    assert dev.target_temperature_cooling == features.TargetTemperatureFeature(
        value=22, step=0.5, min=16, max=30, instance="cooling-target"
    )
    assert dev.numbers == {
        ("timer", None): features.NumbersFeature(
            value=0, min=0, max=1440, step=30, name="timer", unit="minutes"
        )
    }
    assert dev.selects == {
        ("fan-speed", "ac-fan-speed"): features.SelectFeature(
            selected="fan-speed-auto",
            selects={"fan-speed-auto", "fan-speed-2-100", "fan-speed-2-050"},
            name="Fan Speed",
        ),
        ("sleep", None): features.SelectFeature(
            selected="off",
            selects={"on", "off"},
            name="Sleep Mode",
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("portable-ac.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "available",
                "value": False,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 19,
                "lastUpdateTime": 0,
                "functionInstance": "current-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 18,
                "lastUpdateTime": 0,
                "functionInstance": "cooling-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "mode",
                "value": "cool",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            functionClass="timer",
            functionInstance=None,
            lastUpdateTime=0,
            value=60,
        ),
        AferoState(
            functionClass="fan-speed",
            functionInstance="ac-fan-speed",
            lastUpdateTime=0,
            value="fan-speed-2-100",
        ),
        AferoState(
            functionClass="temperature-units",
            functionInstance=None,
            lastUpdateTime=0,
            value="celsius",
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is False
    assert dev.current_temperature == features.CurrentTemperatureFeature(
        temperature=19,
        function_class="temperature",
        function_instance="current-temp",
    )
    assert dev.target_temperature_cooling.value == 18
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.numbers[("timer", None)].value == 60
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
    assert updates == {
        "available",
        "temperature-cooling-target",
        "temperature-current-temp",
        "mode",
        "number-('timer', None)",
        "select-('fan-speed', 'ac-fan-speed')",
        "temperature-units",
    }


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("portable-ac.json")[0]
    updates = await mocked_controller.update_elem(dev_update)
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    mocked_controller[portable_ac_id].display_celsius = True
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=22.5,
        numbers={("timer", None): 60},
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 22.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.numbers[("timer", None)].value == 60
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == portable_ac_id
    expected_calls = [
        {
            "functionClass": "mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "cool",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "cooling-target",
            "lastUpdateTime": 12345,
            "value": 22.5,
        },
        {
            "functionClass": "timer",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": 60,
        },
        {
            "functionClass": "fan-speed",
            "functionInstance": "ac-fan-speed",
            "lastUpdateTime": 12345,
            "value": "fan-speed-2-100",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "current-temp",
            "lastUpdateTime": 12345,
            "value": 36,
        },
    ]
    for call in expected_calls:
        assert call in post["values"]
    assert len(expected_calls) == len(post["values"])


@pytest.mark.asyncio
async def test_set_state_in_f(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=76,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 24.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == portable_ac_id
    expected_calls = [
        {
            "functionClass": "mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "cool",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "cooling-target",
            "lastUpdateTime": 12345,
            "value": 24.5,
        },
        {
            "functionClass": "fan-speed",
            "functionInstance": "ac-fan-speed",
            "lastUpdateTime": 12345,
            "value": "fan-speed-2-100",
        },
    ]
    for call in expected_calls:
        assert call in post["values"]
    assert len(expected_calls) == len(post["values"])


@pytest.mark.asyncio
async def test_set_state_in_f_force_c(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="cool",
        target_temperature=24.5,
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
        is_celsius=True,
    )
    dev = mocked_controller.items[0]
    assert dev.target_temperature_cooling.value == 24.5
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "auto-cool"
    assert dev.selects[("fan-speed", "ac-fan-speed")].selected == "fan-speed-2-100"
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == portable_ac_id
    expected_calls = [
        {
            "functionClass": "mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "cool",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "cooling-target",
            "lastUpdateTime": 12345,
            "value": 24.5,
        },
        {
            "functionClass": "fan-speed",
            "functionInstance": "ac-fan-speed",
            "lastUpdateTime": 12345,
            "value": "fan-speed-2-100",
        },
    ]
    for call in expected_calls:
        assert call in post["values"]
    assert len(expected_calls) == len(post["values"])


@pytest.mark.asyncio
async def test_set_state_invalid_dev(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        "nope",
        hvac_mode="cool",
        target_temperature=22.5,
        numbers={("timer", None): 60},
        selects={("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_set_state_no_updates(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
    )
    mocked_controller._bridge.request.assert_not_called()


@pytest.mark.asyncio
async def test_set_state_invalid_updates(mocked_controller):
    await mocked_controller.initialize_elem(portable_ac)
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        portable_ac_id,
        hvac_mode="i dont exist",
        numbers={(None, None): 7, ("timer", None): 60},
        selects={(None, None): 7, ("fan-speed", "ac-fan-speed"): "fan-speed-2-100"},
    )
    mocked_controller._bridge.request.assert_called_once()
    expected_calls = [
        {
            "functionClass": "timer",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": 60,
        },
        {
            "functionClass": "fan-speed",
            "functionInstance": "ac-fan-speed",
            "lastUpdateTime": 12345,
            "value": "fan-speed-2-100",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "current-temp",
            "lastUpdateTime": 12345,
            "value": 36,
        },
    ]
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == portable_ac_id
    for call in expected_calls:
        assert call in post["values"]
    assert len(expected_calls) == len(post["values"])
